@@ -36,13 +36,14 @@
 
 #define GNOME_IDLE_MONITOR_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GNOME_TYPE_IDLE_MONITOR, GnomeIdleMonitorPrivate))
 
-#define USER_ACTIVE_WATCH_ID 1
+G_STATIC_ASSERT(sizeof(unsigned long) == sizeof(gpointer));
 
 struct _GnomeIdleMonitorPrivate
 {
 	Display	    *display;
 
 	GHashTable  *watches;
+	GHashTable  *alarms;
 	int	     sync_event_base;
 	XSyncCounter counter;
 
@@ -53,7 +54,7 @@ struct _GnomeIdleMonitorPrivate
 
 typedef struct
 {
-	Display			 *display;
+	GnomeIdleMonitor         *monitor;
 	guint			  id;
 	GnomeIdleMonitorWatchFunc callback;
 	gpointer		  user_data;
@@ -70,7 +71,11 @@ enum
 
 static GParamSpec *obj_props[PROP_LAST];
 
-G_DEFINE_TYPE (GnomeIdleMonitor, gnome_idle_monitor, G_TYPE_OBJECT)
+static void gnome_idle_monitor_initable_iface_init (GInitableIface *iface);
+
+G_DEFINE_TYPE_WITH_CODE (GnomeIdleMonitor, gnome_idle_monitor, G_TYPE_OBJECT,
+			 G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE,
+						gnome_idle_monitor_initable_iface_init))
 
 static gint64
 _xsyncvalue_to_int64 (XSyncValue value)
@@ -81,14 +86,6 @@ _xsyncvalue_to_int64 (XSyncValue value)
 
 #define GINT64_TO_XSYNCVALUE(value, ret) XSyncIntsToValue (ret, value, ((guint64)value) >> 32)
 
-static gboolean
-_find_alarm (gpointer		    key,
-	     GnomeIdleMonitorWatch *watch,
-	     XSyncAlarm		   *alarm)
-{
-	return watch->xalarm == *alarm;
-}
-
 static XSyncAlarm
 _xsync_alarm_set (GnomeIdleMonitor	*monitor,
 		  XSyncTestType          test_type,
@@ -135,45 +132,67 @@ set_alarm_enabled (Display    *dpy,
 	XSyncChangeAlarm (dpy, alarm, XSyncCAEvents, &attr);
 }
 
-static GnomeIdleMonitorWatch *
-find_watch_for_alarm (GnomeIdleMonitor *monitor,
-		      XSyncAlarm	alarm)
+static void
+fire_watch (gpointer data,
+	    gpointer user_data)
 {
-	GnomeIdleMonitorWatch *watch;
+	GnomeIdleMonitorWatch *watch = data;
+	XSyncAlarm alarm = (XSyncAlarm) user_data;
 
-	watch = g_hash_table_find (monitor->priv->watches,
-				   (GHRFunc)_find_alarm,
-				   &alarm);
-	return watch;
+	if (watch->xalarm != alarm) {
+		return;
+	}
+
+	g_object_ref (watch->monitor);
+
+	if (watch->callback) {
+		watch->callback (watch->monitor,
+				 watch->id,
+				 watch->user_data);
+	}
+
+	if (watch->xalarm == watch->monitor->priv->user_active_alarm) {
+		gnome_idle_monitor_remove_watch (watch->monitor, watch->id);
+	}
+
+	g_object_unref (watch->monitor);
 }
 
 static void
 handle_alarm_notify_event (GnomeIdleMonitor	    *monitor,
 			   XSyncAlarmNotifyEvent    *alarm_event)
 {
-	GnomeIdleMonitorWatch *watch;
+	XSyncAlarm alarm;
+	GList *watches;
+	gboolean has_alarm;
 
 	if (alarm_event->state != XSyncAlarmActive) {
 		return;
 	}
 
-	watch = find_watch_for_alarm (monitor, alarm_event->alarm);
-	if (watch == NULL)
-		return;
+	alarm = alarm_event->alarm;
 
-	if (watch->id == USER_ACTIVE_WATCH_ID) {
+	has_alarm = FALSE;
+
+	if (alarm == monitor->priv->user_active_alarm) {
 		set_alarm_enabled (monitor->priv->display,
-				   watch->xalarm,
+				   alarm,
 				   FALSE);
-	} else {
+		has_alarm = TRUE;
+	} else if (g_hash_table_contains (monitor->priv->alarms, (gpointer) alarm)) {
 		ensure_alarm_rescheduled (monitor->priv->display,
-					  watch->xalarm);
+					  alarm);
+		has_alarm = TRUE;
 	}
 
-	if (watch->callback != NULL) {
-		watch->callback (monitor,
-				 watch->id,
-				 watch->user_data);
+	if (has_alarm) {
+		watches = g_hash_table_get_values (monitor->priv->watches);
+
+		g_list_foreach (watches,
+				fire_watch,
+				(gpointer) alarm);
+
+		g_list_free (watches);
 	}
 }
 
@@ -234,7 +253,7 @@ find_idletime_counter (GnomeIdleMonitor *monitor)
 static guint32
 get_next_watch_serial (void)
 {
-	static guint32 serial = USER_ACTIVE_WATCH_ID;
+	static guint32 serial = 0;
 	g_atomic_int_inc (&serial);
 	return serial;
 }
@@ -242,17 +261,23 @@ get_next_watch_serial (void)
 static void
 idle_monitor_watch_free (GnomeIdleMonitorWatch *watch)
 {
+	GnomeIdleMonitor *monitor;
+
 	if (watch == NULL) {
 		return;
 	}
 
+	monitor = watch->monitor;
+
 	if (watch->notify != NULL) {
-	    watch->notify (watch->user_data);
+		watch->notify (watch->user_data);
 	}
 
-	if (watch->id != USER_ACTIVE_WATCH_ID) {
-		XSyncDestroyAlarm (watch->display, watch->xalarm);
+	if (watch->xalarm != monitor->priv->user_active_alarm) {
+		XSyncDestroyAlarm (monitor->priv->display, watch->xalarm);
+		g_hash_table_remove (monitor->priv->alarms, (gpointer) watch->xalarm);
 	}
+
 	g_slice_free (GnomeIdleMonitorWatch, watch);
 }
 
@@ -279,10 +304,9 @@ init_xsync (GnomeIdleMonitor *monitor)
 	}
 
 	monitor->priv->counter = find_idletime_counter (monitor);
-	if (monitor->priv->counter == None) {
-		g_warning ("GnomeIdleMonitor: IDLETIME counter not found");
+	/* IDLETIME counter not found? */
+	if (monitor->priv->counter == None)
 		return;
-	}
 
 	monitor->priv->user_active_alarm = _xsync_alarm_set (monitor, XSyncNegativeTransition, 1, FALSE);
 
@@ -295,15 +319,18 @@ gnome_idle_monitor_dispose (GObject *object)
 	GnomeIdleMonitor *monitor;
 
 	monitor = GNOME_IDLE_MONITOR (object);
-	if (monitor->priv->watches != NULL) {
-		g_hash_table_destroy (monitor->priv->watches);
-		monitor->priv->watches = NULL;
+
+	g_clear_pointer (&monitor->priv->watches, g_hash_table_destroy);
+	g_clear_pointer (&monitor->priv->alarms, g_hash_table_destroy);
+	g_clear_object (&monitor->priv->device);
+
+	if (monitor->priv->user_active_alarm != None) {
+		XSyncDestroyAlarm (monitor->priv->display, monitor->priv->user_active_alarm);
+		monitor->priv->user_active_alarm = None;
 	}
 
 	gdk_window_remove_filter (NULL, (GdkFilterFunc)xevent_filter, monitor);
 
-	g_clear_object (&monitor->priv->device);
-
 	G_OBJECT_CLASS (gnome_idle_monitor_parent_class)->dispose (object);
 }
 
@@ -352,6 +379,30 @@ gnome_idle_monitor_constructed (GObject *object)
 	init_xsync (monitor);
 }
 
+static gboolean
+gnome_idle_monitor_initable_init (GInitable     *initable,
+				  GCancellable  *cancellable,
+				  GError       **error)
+{
+	GnomeIdleMonitor *monitor;
+
+	monitor = GNOME_IDLE_MONITOR (initable);
+
+	if (monitor->priv->counter == None) {
+		g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
+				     "Per-device idletime is not supported");
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+static void
+gnome_idle_monitor_initable_iface_init (GInitableIface *iface)
+{
+	iface->init = gnome_idle_monitor_initable_init;
+}
+
 static void
 gnome_idle_monitor_class_init (GnomeIdleMonitorClass *klass)
 {
@@ -387,6 +438,8 @@ gnome_idle_monitor_init (GnomeIdleMonitor *monitor)
 							NULL,
 							NULL,
 							(GDestroyNotify)idle_monitor_watch_free);
+
+	monitor->priv->alarms = g_hash_table_new (NULL, NULL);
 }
 
 /**
@@ -399,7 +452,7 @@ gnome_idle_monitor_init (GnomeIdleMonitor *monitor)
 GnomeIdleMonitor *
 gnome_idle_monitor_new (void)
 {
-	return GNOME_IDLE_MONITOR (g_object_new (GNOME_TYPE_IDLE_MONITOR, NULL));
+	return GNOME_IDLE_MONITOR (g_initable_new (GNOME_TYPE_IDLE_MONITOR, NULL, NULL, NULL));
 }
 
 /**
@@ -407,14 +460,38 @@ gnome_idle_monitor_new (void)
  * @device: A #GdkDevice to get the idle time for.
  *
  * Returns: a new #GnomeIdleMonitor that tracks the device-specific
- * idletime for @device. To track server-global idletime for all
- * devices, use gnome_idle_monitor_new().
+ * idletime for @device. If device-specific idletime is not available,
+ * %NULL is returned, and @error is set. To track server-global
+ * idletime for all devices, use gnome_idle_monitor_new().
  */
 GnomeIdleMonitor *
-gnome_idle_monitor_new_for_device (GdkDevice *device)
+gnome_idle_monitor_new_for_device (GdkDevice  *device)
 {
-	return GNOME_IDLE_MONITOR (g_object_new (GNOME_TYPE_IDLE_MONITOR,
-						 "device", device, NULL));
+	return GNOME_IDLE_MONITOR (g_initable_new (GNOME_TYPE_IDLE_MONITOR, NULL, NULL,
+						   "device", device, NULL));
+}
+
+static GnomeIdleMonitorWatch *
+make_watch (GnomeIdleMonitor          *monitor,
+	    XSyncAlarm                 xalarm,
+	    GnomeIdleMonitorWatchFunc  callback,
+	    gpointer                   user_data,
+	    GDestroyNotify             notify)
+{
+	GnomeIdleMonitorWatch *watch;
+
+	watch = g_slice_new0 (GnomeIdleMonitorWatch);
+	watch->monitor = monitor;
+	watch->id = get_next_watch_serial ();
+	watch->callback = callback;
+	watch->user_data = user_data;
+	watch->notify = notify;
+	watch->xalarm = xalarm;
+
+	g_hash_table_insert (monitor->priv->watches,
+			     GUINT_TO_POINTER (watch->id),
+			     watch);
+	return watch;
 }
 
 /**
@@ -450,17 +527,15 @@ gnome_idle_monitor_add_idle_watch (GnomeIdleMonitor	       *monitor,
 
 	g_return_val_if_fail (GNOME_IS_IDLE_MONITOR (monitor), 0);
 
-	watch = g_slice_new0 (GnomeIdleMonitorWatch);
-	watch->id = get_next_watch_serial ();
-	watch->display = monitor->priv->display;
-	watch->callback = callback;
-	watch->user_data = user_data;
-	watch->notify = notify;
-	watch->xalarm = _xsync_alarm_set (monitor, XSyncPositiveTransition, interval_msec, TRUE);
+	watch = make_watch (monitor,
+			    _xsync_alarm_set (monitor, XSyncPositiveTransition, interval_msec, TRUE),
+			    callback,
+			    user_data,
+			    notify);
+
+	g_hash_table_add (monitor->priv->alarms,
+			  (gpointer) watch->xalarm);
 
-	g_hash_table_insert (monitor->priv->watches,
-			     GUINT_TO_POINTER (watch->id),
-			     watch);
 	return watch->id;
 }
 
@@ -490,21 +565,15 @@ gnome_idle_monitor_add_user_active_watch (GnomeIdleMonitor          *monitor,
 
 	g_return_val_if_fail (GNOME_IS_IDLE_MONITOR (monitor), 0);
 
-	watch = g_slice_new0 (GnomeIdleMonitorWatch);
-	watch->id = USER_ACTIVE_WATCH_ID;
-	watch->display = monitor->priv->display;
-	watch->callback = callback;
-	watch->user_data = user_data;
-	watch->notify = notify;
-	watch->xalarm = monitor->priv->user_active_alarm;
-
 	set_alarm_enabled (monitor->priv->display,
 			   monitor->priv->user_active_alarm,
 			   TRUE);
 
-	g_hash_table_insert (monitor->priv->watches,
-			     GUINT_TO_POINTER (watch->id),
-			     watch);
+	watch = make_watch (monitor,
+			    monitor->priv->user_active_alarm,
+			    callback,
+			    user_data,
+			    notify);
 
 	return watch->id;
 }
